#!feature[nll]

extern crate prusti_contracts;

// Wrapper type for Rust's Vec (instantiated for i32 values)
pub struct VecI32{
    v: Vec<i32>
}

// Wrapper functions around Rust's Vec library:
impl VecI32 {

    #[trusted]
    #[pure]
    #[ensures="result >= 0"]
    pub fn len(&self) -> usize {
        self.v.len()
    }

    #[trusted]
    #[ensures="result.len() == 0"]    
    pub fn with_capacity(capacity: usize) -> Self {
        VecI32{ v: Vec::with_capacity(capacity) }
    }

    #[trusted]
    #[requires="0 <= from && from <= to && to <= origin.len()"]
    #[ensures="result.len() == to-from"]
    #[ensures="forall i: usize :: (0 <= i && i < result.len()) ==>
    result.lookup(i) == old(origin.lookup(i+from))"]
    pub fn clone_vec_part(origin: &VecI32, from: usize, to: usize) -> VecI32 {
        let mut result = VecI32::with_capacity(to-from);
        let slice = &origin.v[from..to];
        result.v.extend_from_slice(slice);
        result
    }

    #[trusted]
    #[pure]
    #[requires="0 <= index && index < self.len()"]
    pub fn lookup(&self, index: usize) -> i32 {
        self.v[index]
    }

    #[trusted]
    #[requires="0 <= index && index < self.len()"]
    #[ensures="self.len() == old(self.len())"]
    #[ensures="self.lookup(index) == value"]
    #[ensures="forall i: usize :: (0 <= i && i < self.len() && i != index) ==>
                    self.lookup(i) == old(self.lookup(i))"]
    pub fn store(&mut self, index: usize, value: i32) {
        self.v[index] = value;
    }
}

// dummy main function (for compilation)
fn main() {
}

// 1. LENGTH-PRESERVED:
#[ensures="arr.len() == old(arr.len())"]
// 2. SORTEDNESS:
#[ensures="forall i:usize, j:usize :: (0<=i && i<j && j<arr.len()) ==> arr.lookup(i) <= arr.lookup(j)"]
fn merge_sort(arr: &mut VecI32) {
  let l = arr.len(); // BUG: Prusti crashes if we merge these lines
  merge_sort_part(arr,0,l);
}

// Precondition: what do we need about from and until?
#[requires="true"]
// Postcondition 1: what do we need for LENGTH-PRESERVED above?
#[ensures="true"]
// Postcondition 2: Express that the range from "from" up until "until" is sorted
#[ensures="true"]
// Postcondition 3: What about the rest of the vector elements?
#[ensures="true"]
fn merge_sort_part(arr: &mut VecI32, from: usize, until: usize) {
  // Merge Sort Steps
  // 1. split range into two halves (if possible)
    // 2. recursively sort the two halves
    // 3. make a copy of the first half (hint: use the function VecI32::clone_vec_part defined above)
    // 4. merge: walk the copy and second half, writing their elements into the original in order
    // (hint: be careful about "falling off" the ends of the two halves; at some point one will be done)
    // 5. your loop will need invariants! Properties regarding:
    // (a) ranges of index variables
    // (b) relating indices reading from the two halves with index writing into the array
    // (c) arr length unchanged
    // (d) portions still to be read from
    // (e) portion written so far
    // (f) relating last element written to next elements to be read from each half
    // (g) stating that arr values outside of indices [from..until) is unchanged
    // Note: a current bug means some specifications of the form b1 && b2 ==> b3 will not be parsed 
    // correctly. If you encounter this issue, try writing this as b1 ==> (b2 ==> b3) instead.
}
