#![feature(box_syntax, box_patterns)]
extern crate prusti_contracts;

struct Point {
  x: i32, y: i32
}

// Postconditions needed: what changes and how? What does not?
#[ensures="p.x == old(p.x) + s"]
#[ensures="p.y == old(p.y)"]
fn shift_x(p: &mut Point, s: i32) {
  p.x = p.x + s
}

struct Route { //list of Points
  current: Point,
  rest: Option<Box<Route>>
}

#[pure]
// Postcondition needed: what do we know about result (in all cases)?
#[ensures="result > 0"]
fn len(r: &Route) -> i32 {
  1 + match r.rest {
    Some(box ref q) => len(q),
    None => 0
  }
}

#[pure]
// Precondition needed: what do we require about n?
#[requires="0 <= n && n < len(r)"]
fn get_nth_x(r: &Route, n: i32) -> i32 {
  if n == 0 { r.current.x } else {
    match r.rest {
      Some(box ref r) => get_nth_x(r, n-1),
      None => unreachable!()
    }
  }
}


// Precondition: what is required about n?
#[requires="0 <= n && n < len(r)"]
// Postcondition: len doesn't change
#[ensures="len(r) == old(len(r))"]
// Postcondition: one point's x coordinate has changed
#[ensures="get_nth_x(r, n) ==
  old(get_nth_x(r, n)) + s"]
// Postcondition: other points' x coordinates have not changed
#[ensures="forall i: i32 ::
  (0<=i && i<len(r) && i != n) ==>
  get_nth_x(r, i) == old(get_nth_x(r, i))"]
fn shift_nth_x_versionA(r: &mut Route, n: i32, s:i32) {
  if n == 0 { 
    shift_x(&mut r.current, s) 
  } else {
    match r.rest {
      Some(box ref mut rest) =>
        shift_nth_x_versionA(rest, n-1, s),
      None => unreachable!()
    }
  }
}

// TASK A: fill in appropriate specifications for the functions above this line, such that they verify

// TASK B: uncomment the function below, and add an appropriate specification to it.
// Then, reimplement shift_nth_x by simply calling this function and the shift_x function appropriately.
// You should not need to change the specification of shift_nth_x; only its body.
// Note that this means that the refactoring is opaque for the verification of callers of shift_nth_x

// Precondition: n values
#[requires="0 <= n && n < len(r)"]
// Postcondition: what is the *current* x coordinate of result?
#[ensures="result.x == old(get_nth_x(r, n))"]
// Postcondition: what *will* the x coordinates of the points in r be *after* the result expires?
// Hint: you need a pledge to express this.
#[ensures="after_expiry(
    len(r) == old(len(r)) &&
    get_nth_x(r, n) == before_expiry(result.x) &&
    forall i: i32 ::
        (0<=i && i<len(r) && i != n) ==>
        get_nth_x(r, i) == old(get_nth_x(r, i))
)
"]
fn borrow_nth(r:&mut Route, n: i32) -> &mut Point {
  if n == 0 { &mut r.current } else {
    match r.rest {
      Some(box ref mut r) =>
        borrow_nth(r, n-1),
      None => unreachable!()
    }
  }
}

#[requires="0 <= n && n < len(r)"]
#[ensures="len(r) == old(len(r))"]
#[ensures="get_nth_x(r, n) ==
  old(get_nth_x(r, n)) + s"]
#[ensures="forall i: i32 ::
  (0<=i && i<len(r) && i != n) ==>
  get_nth_x(r, i) == old(get_nth_x(r, i))"]
fn shift_nth_x(r: &mut Route, n: i32, s:i32) {
  let p = borrow_nth(r, n);
  shift_x(p,s);
}





// dummy main function (for compilation)
fn main() {}
